#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "macros.h"
#include "f2c.h"
#include "clapack.h"

#include "linalg.h"

/* code for handling CLAPACK errors as reported in the INFO variable.
   the error reporting is identical for dgetrf and dgetri */
#define my_handle_errors(info, fcn)				  \
  if(info != 0) {						  \
    if(info < 0) { /* invalid argument */			  \
      DIE(#fcn "_() reports invalid arg %d\n", (int)-info);	  \
    } else {							  \
      printf("singular matrix\n");                                \
      print_colmaj_matrix_dec(A, n, n);                           \
      DIE(#fcn "_() reports singular matrix with U(%d,%d) = 0\n", \
	  (int)info, (int)info);				  \
      assert(0);                                                  \
    }								  \
  }



/* Compute the inverse of a matrix A via the LU factorization, using
   the same factorization to compute the log of the *absolute value*
   of the determinant.  NB: this will overwrite A with A^-1 */
double det_inv(double* A, integer LDA, integer n) {
  int i;

  assert(A != NULL); assert(n > 0); assert(LDA >= n);

  /* static storage for P (row-permuatation) array */
  static integer* P = NULL;
  static int P_size = 0;
  if(P_size < n) {
    P = (integer*)realloc(P, (P_size = n)*sizeof(integer));
    assert(P != NULL);
  }

  /* static storage for WORK (workspace for inversion)*/
  static double* WORK = NULL;
  static int WORK_size = 0;

  integer INFO = 0;

  /* perform LU factorization (overwritten on A) */
  dgetrf_(&n, &n, A, &LDA, P, &INFO);
  my_handle_errors(INFO, dgetrf);

  /* compute determinant as product of diagonal of U (which is stored
     as the upper triangle of A) */
  double logdet = 0;
  const int i_max = n*LDA;
  const int i_inc = LDA + 1;
  for(i = 0; i < i_max; i += i_inc) {
    const double A_i = A[i];
    assert(isfinite(A_i));
    logdet += log(A_i < 0 ? -A_i : A_i);  /* absolute value */
  }

  /* begin by making a workspace query */
  double optimal_LWORK = -1;
  integer LWORK = -1;  /* indicator for workspace query */
  dgetri_(&n, A, &LDA, P, &optimal_LWORK, &LWORK, &INFO);
  my_handle_errors(INFO, dgetri);

  /* check that optimal_LWORK is a positive integer */
  assert(optimal_LWORK > 0);
  assert(((double)((int)optimal_LWORK)) == optimal_LWORK);

  /* allocate WORK and perform the inversion */
  LWORK = (integer)optimal_LWORK;
  if(WORK_size < LWORK) {
    WORK = (double*)realloc(WORK, (WORK_size = LWORK)*sizeof(double));
    assert(WORK != NULL);
  }
  dgetri_(&n, A, &LDA, P, WORK, &LWORK, &INFO);
  my_handle_errors(INFO, dgetri);

  return logdet;
}

#undef my_handle_errors /* clean up */
