#ifndef LINALG_H
#define LINALG_H

#include <stdio.h> /* for FILE below */

#include "f2c.h"   /* for integer */

/* there will inevitably be numerical errors that cause some of the
   marginal probabilities that would ordinarily be close to zero to
   change sign and be very slightly negative.  this messes up the
   log-marginals, so we use the following to floor all probabilities
   at positive values. */
#define MY_EPSILON ((double)1e-300)
#define force_positive(x) (x < MY_EPSILON ? MY_EPSILON : x)

/* threshhold for differences in log-values */
#define LOG_DYNAMIC_RANGE 20

/* ensure that numerical errors in negative marginals do not go beyond
   a certain value */
#define MY_TOLERANCE ((double)1e-30)
#define check_marginal(m, i, j)						\
  if(m < -MY_TOLERANCE) {						\
    printf(__FILE__							\
	   ":%d: negative marginal (%.15g) exceeded "			\
	   "tolerance for numerical errors (%.15g) "			\
	   "at position %d,%d\n",					\
	   __LINE__, m, MY_TOLERANCE, i, j);				\
  } else if(!isfinite(m)) {						\
    printf(__FILE__							\
	   ":%d: non-finite marginal (%.15g) at position %d,%d\n",	\
	   __LINE__, m, i, j);						\
  }


/* Compute the inverse of a matrix A via the LU factorization, using
   the same factorization to compute the log of the *absolute value*
   of the determinant.  NB: this will overwrite A with A^-1 */
double det_inv(double* A, integer LDA, integer n);
/* Compute the log of the *absolute value* of the determinant.  NB:
   this will overwrite A with its LU decomposition */
double det(double* A, integer LDA, integer n);


/* create a random matrix with the given dimensions.  memory is
   allocated with malloc() and can be freed with free() */
double* rand_matrix(const int m, const int n);

/* create a mallocated copy of the given matrix */
double* copy_matrix(const double* const restrict a, const int m, const int n);

/* read a matrix A from a file with the following format:
   <m> <n> <A_1,1> ... <A_1,n> <A_2,1> ... <A_m,1> ... <A_m,n>
   where arbitrary whitespace can intervene */
double* read_matrix(FILE* const f, int* const ret_m, int* const ret_n);
double* read_sized_matrix(FILE* const f, const int m, const int n);

void write_matrix(FILE* const f, const double* const restrict a,
		  const int m, const int n);

/* print the rows of a matrix in column-major format */
void print_colmaj_matrix(const double* const restrict a,
			 const int m, const int n);
void fprint_colmaj_matrix(FILE* const f, const double* const restrict a,
			  const int m, const int n);

/* same as above but in decimal format */
void write_matrix_dec(FILE* const f, const double* const restrict a,
		      const int m, const int n);
void print_colmaj_matrix_dec(const double* const restrict a,
			     const int m, const int n);
void fprint_colmaj_matrix_dec(FILE* const f, const double* const restrict a,
			      const int m, const int n);

/* perform the multiplication C = A'B */
void dgemm_C_eq_At_B(double* C, const int LDC,
		     const int Cm, const int Cn,
		     const double* A, const int LDA,
		     const int Am, const int An,
		     const double* B, const int LDB, 
		     const int Bm, const int Bn);

/* perform the multiplication C = AB */
void dgemm_C_eq_A_B(double* C, const int LDC,
		    const int Cm, const int Cn,
		    const double* A, const int LDA,
		    const int Am, const int An,
		    const double* B, const int LDB, 
		    const int Bm, const int Bn);

#endif /* LINALG_H */
