#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <assert.h>
#include <string.h>

#include "cblas.h"
#include "linalg.h"
#include "f2c.h"
#include "clapack.h"
#include "macros.h"

double* rand_matrix(const int m, const int n) {
  int i;
  assert(m > 0);  assert(n > 0);
  /* check for integer overflow */
  const double dsize = (double)m * (double)n * (double)sizeof(double);
  assert(dsize == (size_t)dsize);
  double* const restrict a = (double*)malloc((size_t)dsize);
  assert(a != NULL);
  const size_t size = (size_t)m * (size_t)n;
  for(i = 0; i < size; ++i) {
    a[i] = drand();
  }
  return a;
}

double* copy_matrix(const double* const restrict a, const int m, const int n) {
  assert(m > 0);  assert(n > 0);
  int i;
  /* check for integer overflow */
  const double dsize = (double)m * (double)n * (double)sizeof(double);
  assert(dsize == (size_t)dsize);
  double* const restrict b = (double*)malloc((size_t)dsize);
  const size_t size = (size_t)m * (size_t)n;
  for(i = 0; i < size; ++i) { b[i] = a[i]; }
  return b;
}

double* read_matrix(FILE* const f, int* const ret_m, int* const ret_n) {
  int m, n;
  assert(f != NULL);  assert(!feof(f));
  assert(fscanf(f, "%d %d", &m, &n) == 2);
  assert(m > 0);  assert(n > 0);
  *ret_m = m;  *ret_n = n;
  return read_sized_matrix(f, m, n);
}

double* read_sized_matrix(FILE* const f, const int m, const int n) {
  int i, idx;
  assert(f != NULL);  assert(!feof(f));
  assert(m > 0);  assert(n > 0);
  /* check for integer overflow */
  const double dsize = (double)m * (double)n * (double)sizeof(double);
  assert(dsize == (size_t)dsize);
  double* const restrict a = (double*)malloc((size_t)dsize);
  assert(a != NULL);
  /* these paired for loops iterate `idx' through a row-major walk in
     a in column-major matrix */
  const size_t size = (size_t)m * (size_t)n;
  for(i = 0; i < m; ++i) {
    for(idx = i; idx < size; idx += m) {
      assert(fscanf(f, " %lf ", (a + idx)) == 1);
    }
  }
  return a;
}

void write_matrix(FILE* const f, const double* const restrict a,
		  const int m, const int n) {
  fprintf(f, "%d %d\n", m, n);
  fprint_colmaj_matrix(f, a, m, n);
}


void print_colmaj_matrix(const double* const restrict a,
			 const int m, const int n) {
  fprint_colmaj_matrix(stdout, a, m, n);
}

void fprint_colmaj_matrix(FILE* const f, const double* const restrict a,
			  const int m, const int n) {
  int i, j;
  assert(m >= 0);  assert(n >= 0);
  if(m == 0 || n == 0) { return; }
  assert(a != NULL);
  for(i = 0; i < m; ++i) {
    for(j = 0; j < n; ++j) {
      const int tail = (j == n-1 ? '\n' : ' ');
      const double val = a[m*j + i];
      if(val == 0) {
	fprintf(f, "0%c", tail);
      } else {
	fprintf(f, "%a%c", val, tail);
      }
    }
  }
}


void write_matrix_dec(FILE* const f, const double* const restrict a,
		      const int m, const int n) {
  fprintf(f, "%d %d\n", m, n);
  fprint_colmaj_matrix_dec(f, a, m, n);
}


void print_colmaj_matrix_dec(const double* const restrict a,
			     const int m, const int n) {
  fprint_colmaj_matrix_dec(stdout, a, m, n);
}

void fprint_colmaj_matrix_dec(FILE* const f, const double* const restrict a,
			      const int m, const int n) {
  int i, j;
  assert(m >= 0);  assert(n >= 0);
  if(m == 0 || n == 0) { return; }
  assert(a != NULL);
  for(i = 0; i < m; ++i) {
    for(j = 0; j < n; ++j) {
      const int tail = (j == n-1 ? '\n' : ' ');
      const double val = a[m*j + i];
      if(val == 0) {
	fprintf(f, "0%c", tail);
      } else {
	fprintf(f, "%.15g%c", val, tail);
      }
    }
  }
}


/* perform the multiplication C = A'B */
void dgemm_C_eq_At_B(double* C, const int LDC,
		     const int Cm, const int Cn,
		     const double* A, const int LDA,
		     const int Am, const int An,
		     const double* B, const int LDB, 
		     const int Bm, const int Bn) {
  assert(Cm == An);
  assert(Cn == Bn);
  assert(Am == Bm);
  assert(LDC >= Cm);
  assert(LDA >= Am);
  assert(LDB >= Bm);

  cblas_dgemm(CblasColMajor,  /* treat matrices as col-major */
	      CblasTrans,     /* A' */
	      CblasNoTrans,   /* B */
	      Cm, Cn, Am,     /* dimensions */
	      1.0,            /* no scaling */
	      A, LDA, B, LDB, /* inputs and LD */
	      0.0,            /* no additive shift */
	      C, LDC);        /* output and LD */
}

/* perform the multiplication C = AB */
void dgemm_C_eq_A_B(double* C, const int LDC,
		    const int Cm, const int Cn,
		    const double* A, const int LDA,
		    const int Am, const int An,
		    const double* B, const int LDB, 
		    const int Bm, const int Bn) {
  assert(Cm == Am);
  assert(Cn == Bn);
  assert(An == Bm);
  assert(LDC >= Cm);
  assert(LDA >= Am);
  assert(LDB >= Bm);

  cblas_dgemm(CblasColMajor,  /* treat matrices as col-major */
	      CblasNoTrans,   /* A */
	      CblasNoTrans,   /* B */
	      Cm, Cn, An,     /* dimensions */
	      1.0,            /* no scaling */
	      A, LDA, B, LDB, /* inputs and LD */
	      0.0,            /* no additive shift */
	      C, LDC);        /* output and LD */
}
