#ifndef MATRIX_TREE_H
#define MATRIX_TREE_H

/* matrixtree.h

   Routines for computing the partition function and marginals of a
   (non-negative) weighted directed graph, with respect to the space
   of spanning trees OR spanning forests.  The key algorithmic
   insights are:

   * The use of the Matrix-Tree Theorem (Kirchoff, Tutte) to compute
     the partition function (plus some additional tricks to
     incorporate root-selection scores).

   * Computing the marginals via the derivative of the log-partition
     function, which has a simple closed form as an inverse matrix.

   Cf. Koo, Globerson, Carreras and Collins (EMNLP 2007) for more on
   the theoretical underpinnings (and please cite this work if you
   make use of this package).  In the terminology of the above paper,
   a "spanning tree" is a "single-root non-projective dependency
   structure", and a "spanning forest" is a "multi-root non-projective
   dependency structure".

   Notes on usage:

   * Matrices are assumed to be column-major.  Specifically, an n x n
     matrix A is represented as an array where A[0,...,n-1] contains
     the first column of A, A[n,...,2n-1] contains the second column,
     etc.

   * An adjacency matrix A has the following form: the h'th row gives
     the scores of dependencies with head word h, and the m'th column
     gives the scores of dependencies which have modifier m.  Thus,
     A_{h,m} gives the score of the dependency (h --> m).  The
     diagonal entries of the adjacency matrix give the scores of the
     root dependencies; i.e., A_{m,m} gives the score of word m being
     a child of the root-token.
 */


/* Convert an adjacency matrix to the graph Laplacian of the implied
   graph.

    A : input/output.  on entry, contains the adjacency matrix.  on
        exit, contains the graph Laplacian
    n : input.  #nodes in the graph. */
void make_laplacian(double* const restrict A, const int n);

/* Compute the spanning-tree log-partition function and edge/root
   marginals, given a graph Laplacian and root scores.

   A : input/output.  on entry, contains an adjacency matrix, with
       root-selection scores on the diagonal.  on exit, contains the
       edge (off-diagonal) and root (diagonal) marginals.
   n : input.  #nodes in the graph. */
double st_par_mar(double* const restrict A, const int n);

/* Compute the spanning-tree log-partition function only, given a
   graph Laplacian and root scores.  The partition function may differ
   very slightly from the partition function as found by st_par_mar().

   A : input.  on entry, contains an adjacency matrix, with
       root-selection scores on the diagonal.  NB: even though A is
       not used as a return value, it will be overwritten.
   n : input.  #nodes in the graph. */
double st_par(double* const restrict A, const int n);

/* Compute the spanning-forest log-partition function and edge/root
   marginals, given a graph Laplacian and root scores.

   A : input/output.  on entry, contains an adjacency matrix, with
       root-selection scores on the diagonal.  on exit, contains the
       edge (off-diagonal) and root (diagonal) marginals.
   n : input.  #nodes in the graph. */
double sf_par_mar(double* const restrict A, const int n);

/* Compute the spanning-forest log-partition function only, given a
   graph Laplacian and root scores.  The partition function may differ
   very slightly from the partition function as found by sf_par_mar().

   A : input.  on entry, contains an adjacency matrix, with
       root-selection scores on the diagonal.  NB: even though A is
       not used as a return value, it will be overwritten.
   n : input.  #nodes in the graph. */
double sf_par(double* const restrict A, const int n);

#endif /* MATRIX_TREE_H */
