#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "linalg.h"      /* for det() */
#include "matrixtree.h"  /* for make_laplacian() */

double sf_par(double* const restrict A, const int n) {
  assert(A != NULL); assert(n > 0);
  const int A_size = n*n;
  assert(A[A_size - 1] == A[A_size - 1]);  /* test array bounds for segfault */

  int j;

  /* compensate for numerical errors by scaling each column (i.e.,
     each group of in-edges) so that the maximum value is 1.  this
     will rescale the determinant by the product of all the scaling
     factors, and should not affect the marginals.  Simultaneously
     exponentiate the log-weights. */
  double logpartition_adjust = 0;
  for(j = 0; j < A_size; j += n) {
    const int col_end = j + n;
    int ij;
    /* find the max log-weight */
    double max_weight = A[j];
    for(ij = j + 1; ij < col_end; ++ij) {
      const double weight = A[ij];
      if(weight > max_weight) { max_weight = weight; }
    }
    /* rescale the row and accumulate the scaling factor */
    for(ij = j; ij < col_end; ++ij) {
      A[ij] = exp(A[ij] - max_weight);
    }
    logpartition_adjust += max_weight;
  }

  /* now, we create a graph laplacian in A, adding the root scores on
     the diagonal */
  for(j = 0; j < n; ++j) {
    const int col_start = j*n;
    const int jj = col_start + j;
    const int col_end = col_start + n;
    int ij;
    double sum = 0;
    for(ij = col_start; ij < col_end; ++ij) {
      if(ij != jj) { /* ignore diagonal entries */
	const double A_ij = A[ij];
	sum += A_ij;    /* accumulate in-weight */
	/* Laplacian has negative edge weights */
	A[ij] = -A_ij;
      }
    }
    /* diagonal of Laplacian is in-weight *plus* root score (which is
       already stored on diagonal) */
    A[jj] += sum;
  }

  /* recover the log-partition function and derivative (inverse) */
  const double logdet = det(A, n, n);
  assert(isfinite(logdet));
  const double logpartition = logdet + logpartition_adjust;
  assert(isfinite(logpartition));

  return logpartition;
}
