#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "linalg.h"      /* for det_inv() */
#include "matrixtree.h"  /* for make_laplacian() */

double sf_par_mar(double* const restrict A, const int n) {
  assert(A != NULL); assert(n > 0);
  const int A_size = n*n;
  assert(A[A_size - 1] == A[A_size - 1]);  /* test array bounds for segfault */

  /* static storage for dL (inverse matrix) */
  static double* restrict dL = NULL;
  static int dL_size = 0;
  if(A_size > dL_size) {
    dL = (double*)realloc(dL, (dL_size = A_size)*sizeof(double));
    assert(dL != NULL);
  }

  int j;

  /* compensate for numerical errors by scaling each column (i.e.,
     each group of in-edges) so that the maximum value is 1.  this
     will rescale the determinant by the product of all the scaling
     factors, and should not affect the marginals.  Simultaneously
     exponentiate the log-weights. */
  double logpartition_adjust = 0;
  for(j = 0; j < A_size; j += n) {
    const int col_end = j + n;
    int ij;
    /* find the max log-weight */
    double max_weight = A[j];
    for(ij = j + 1; ij < col_end; ++ij) {
      const double weight = A[ij];
      if(weight > max_weight) { max_weight = weight; }
    }
    /* rescale the row and accumulate the scaling factor */
    for(ij = j; ij < col_end; ++ij) {
      const double norm_weight = A[ij] - max_weight;
      /* NB: the code tends to accumulate numerical errors if the
	 differences in log-weight have a large dynamic range.  So we
	 cap the differences. */
      A[ij] = exp(norm_weight < -LOG_DYNAMIC_RANGE
		  ? -LOG_DYNAMIC_RANGE : norm_weight);
    }
    logpartition_adjust += max_weight;
  }

  /* Now, we create a transposed Laplacian in the remaining columns of
     dL (note that the derivative of log det(L) is (L^-1)^T =
     (L^T)^-1).  In addition, we add the root scores to the diagonal
     of the Laplacian. */
  for(j = 0; j < n; ++j) {
    const int col_start = j*n;
    const int col_end = col_start + n;
    int ij, ji;   /* indices for the i,j and j,i positions */
    double sum = 0;
    for(ij = col_start, ji = j;
	ij < col_end;
	++ij,           ji += n) {
      if(ij != ji) { /* ignore the diagonal here */
	const double A_ij = A[ij];
	sum += A_ij;    /* accumulate in-weight */
	/* Laplacian has negative edge weights */
	dL[ji] = -A_ij;
      }
    }
    /* diagonal of Laplacian is in-weight */
    const int jj = col_start + j;
    dL[jj] = sum + A[jj];
  }

  /* recover the log-partition function and derivative (inverse) */
  const double det = det_inv(dL, n, n);
  assert(isfinite(det));
  const double logpartition = det + logpartition_adjust;
  assert(isfinite(logpartition));

  /* a bit more work is required to compute the marginals:

     M_ij = A_ij*(dL_jj - dL_ij)
        (since each A_ij appears in L_jj and L_ij)

     M_jj = 1 - \sum_k M_kj
        (since each column must sum to one; i.e., each modifier has a
        parent or is a root)
 */
  for(j = 0; j < n; ++j) {
    const int col_start = j*n;
    const int col_end = col_start + n;
    const int jj = col_start + j;   /* index of diagonal j,j entry */
    const double dL_jj = dL[jj];
    int ij;
    /* set the non-diagonal marginals and accumulate the sum */
    double sum = 0;
    for(ij = col_start; ij < col_end; ++ij) {
      if(ij != jj) {   /* ignore the diagonal (see below) */
	const double A_ij  = A[ij];
	const double dL_ij = dL[ij];
	const double M_ij  = A_ij*(dL_jj - dL_ij);
	//	check_marginal(M_ij, ij - col_start, j);
	sum += (A[ij] = force_positive(M_ij));
      }
    }
    /* handle root marginal */
    const double M_jj = A[jj] * dL[jj];
    //    check_marginal(M_jj, j, j);
    sum += (A[jj] = force_positive(M_jj));
    //    if(sum > 1) {
    //      printf("sum > 1 = %.15g\n", sum);
    //      for(ij = col_start; ij < col_end; ++ij) {
    //	printf("bad column[%d] = %.15g\n", ij - col_start, A[ij]);
    //      }
    //    }
    /* re-normalize the marginals */
    for(ij = col_start; ij < col_end; ++ij) { A[ij] /= sum; }
  }

  return logpartition;
}
