#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "linalg.h"      /* for det_inv() */
#include "matrixtree.h"  /* for make_laplacian() */

double st_par_mar(double* const restrict A, const int n) {
  assert(A != NULL); assert(n > 0);
  const int A_size = n*n;
  assert(A[A_size - 1] == A[A_size - 1]);  /* test array bounds for segfault */

  /* static storage for dL (inverse matrix) */
  static double* restrict dL = NULL;
  static int dL_size = 0;
  if(A_size > dL_size) {
    dL = (double*)realloc(dL, (dL_size = A_size)*sizeof(double));
    assert(dL != NULL);
  }

  int i, j;

  /* compensate for numerical errors by scaling each column (i.e.,
     each group of in-edges) so that the maximum value is 1.  this
     will rescale the determinant by the product of all the scaling
     factors, and should not affect the marginals.  Simultaneously
     exponentiate the log-weights. */
  double logpartition_adjust = 0;
  for(j = 0; j < A_size; j += n) {
    const int col_end = j + n;
    int ij;
    /* find the max log-weight */
    double max_weight = A[j];
    for(ij = j + 1; ij < col_end; ++ij) {
      const double weight = A[ij];
      if(weight > max_weight) { max_weight = weight; }
    }
    /* rescale the row and accumulate the scaling factor */
    for(ij = j; ij < col_end; ++ij) {
      const double norm_weight = A[ij] - max_weight;
      /* NB: the code tends to accumulate numerical errors if the
	 differences in log-weight have a large dynamic range.  So we
	 cap the differences. */
      A[ij] = exp(norm_weight < -LOG_DYNAMIC_RANGE
		  ? -LOG_DYNAMIC_RANGE : norm_weight);
    }
    logpartition_adjust += max_weight;
  }

  /* first, we need to recover the root scores from the diagonal of
     the adjacency matrix and create a graph Laplacian */
  const int diag_step = n + 1; /* step size to traverse diagonal */
  for(i = 0, j = 0; /* i indexes dL(:,1), j indexes diag(A)*/
      i < n;
      ++i,   j += diag_step) {
    dL[i] = A[j]; /* save root scores in first column of dL */
  }

  /* Now, we create a transposed Laplacian in the remaining columns of
     dL (note that the derivative of log det(L) is (L^-1)^T =
     (L^T)^-1). */
  for(j = 0; j < n; ++j) {
    const int col_start = j*n;
    const int col_end = col_start + n;
    int ij, ji;   /* indices for the i,j and j,i positions */
    /* NB: we ignore the first row of L (equivalently, the first
       column of dL), because we have already placed the root scores
       there */
    double sum = 0;
    for(ij = col_start, ji = j;
	ij < col_end;
	++ij,           ji += n) {
      if(ij != ji) { /* ignore diagonal entries */
	const double A_ij = A[ij];
	sum += A_ij;    /* accumulate in-weight */
	/* Laplacian has negative edge weights */
	if(ij > col_start) { dL[ji] = -A_ij; }
      }
    }
    /* diagonal of Laplacian is in-weight */
    if(j > 0) { dL[col_start + j] = sum; }
  }

  /* recover the log-partition function and derivative (inverse) */
  const double det = det_inv(dL, n, n);
  assert(isfinite(det));
  const double logpartition = det + logpartition_adjust;
  assert(isfinite(logpartition));

  /* a bit more work is required to compute the marginals:

     M_ii = A_ii * dL_1i
        (because the first row of dL corresponds to the root weights)
     M_i1 = A_i1 * -dL_i1          for i > 1
        (because there is no diagonal sum for the first column)
     M_1j = A_1j * dL_jj          for j > 1
        (because A_1j only appears in the diagonal sums)
     M_ij = A_ij*(dL_jj - dL_ij)   for i,j > 1
        (this is the typical formula)
 */

  /* handle root marginals and 1 -> j marginals */
  const double M_11 = A[0]*dL[0];
  //    check_marginal(M_11, 0, 0);
  A[0] = force_positive(M_11);
  for(i = n,      j = diag_step;  /* i traverses row 1, j traverses diagonal */
      i < A_size;
      i += n,     j += diag_step) {
    const double M_ii = A[j]*dL[i]; /* M_ii = A_ii * dL_1i */
    //    check_marginal(M_ii, i/n, i/n);
    A[j] = force_positive(M_ii);

    const double M_1j = A[i]*dL[j]; /* M_1j = A_1j * dL_jj */
    //    check_marginal(M_1j, 1, i/n);
    A[i] = force_positive(M_1j);
  }

  /* handle i -> 1 marginals */
  double sum = A[0]; /* initialize with root marginal */
  for(i = 1; i < n; ++i) {
    const double M_i1 = -A[i]*dL[i]; /* M_i1 = A_i1 * -dL_i1 */
    //    check_marginal(M_i1, i, 1);
    sum += (A[i] = force_positive(M_i1));
  }
  /* renormalize first column */
  for(i = 0; i < n; ++i) { A[i] /= sum; }

  /* handle rest of marginals and normalize */
  for(j = 1; j < n; ++j) {
    const int col_start = j*n;
    const int col_end = col_start + n;
    const int jj = col_start + j;   /* index of diagonal j,j entry */
    const double dL_jj = dL[jj];
    int ij;
    /* initialize to include the root and first row marginals, which
       are computed above */
    sum = A[col_start] + A[jj];
    /* NB: handle only i > 1 in the loop below */
    for(ij = col_start + 1; ij < col_end; ++ij) {
      if(ij != jj) {   /* ignore the diagonal (see below) */
	const double A_ij  = A[ij];
	const double dL_ij = dL[ij];
	const double M_ij  = A_ij*(dL_jj - dL_ij);
	//	check_marginal(M_ij, ij - col_start, j);
	sum += (A[ij] = force_positive(M_ij));
      }
    }
    /* renormalize this column.  NB: the first row entry also needs to
       be normalized */
    for(ij = col_start; ij < col_end; ++ij) { A[ij] /= sum; }
  }

  return logpartition;
}
