#ifndef MACROS_H
#define MACROS_H

/* this pair of macros exploits the argument prescan in macro
   expansion to allow stringification of the *value* of a macro.
   consider the following:
   #define foo 3
   macro_tostring2(foo) ==> #foo ==> "foo"
   this happens because no expansion is done on arguments which get
   stringified.  however, by wrapping macro_tostring2() in another
   macro, we can get the outer macro to perform the expansion while
   the inner macro does the stringification:
   macro_tostring(foo) ==> macro_tostring2(3) ==> #3 ==> "3" */
#define macro_tostring(x) macro_tostring2(x)
#define macro_tostring2(x) #x

/* WARN and ERR use a GNU C extension to variadic macros that allows
   special behavior when the variable-length argument is empty.  the
   ## operator takes on a special meaning between a comma `,' and the
   macro `__VA_ARGS__', and when __VA_ARGS__ is empty, the comma is
   deleted (otherwise the result would be malformed C code). */
#ifdef __GNUC__
#define MESG(str, ...) \
  fprintf(stdout, __FILE__ ":" macro_tostring(__LINE__) " %s(): " str "\n", __func__ , ##__VA_ARGS__)
#define WARN(str, ...) \
  fprintf(stdout, "* warning in %s() at " __FILE__ " line " macro_tostring(__LINE__) ": " str "\n", __func__ , ##__VA_ARGS__)
#define ERR(str, ...) \
  fprintf(stdout, "*** error in %s() at " __FILE__ " line " macro_tostring(__LINE__) ": " str "\n", __func__ , ##__VA_ARGS__)
#else /* __GNUC__ */
#define MESG(str, ...) \
  fprintf(stdout, __FILE__ ":" macro_tostring(__LINE__) " %s(): " str "\n", __func__, __VA_ARGS__)
#define WARN(str, ...) \
  fprintf(stdout, "* warning in %s() at " __FILE__ " line " macro_tostring(__LINE__) ": " str "\n", __func__, __VA_ARGS__)
#define ERR(str, ...) \
  fprintf(stdout, "*** error in %s() at " __FILE__ " line " macro_tostring(__LINE__) ": " str "\n", __func__, __VA_ARGS__)
#endif /* __GNUC__ */

/* this is basically an analogue to the PERL `die' function */
#define DIE(...) \
  do { \
    ERR(__VA_ARGS__); \
    exit(1); \
  } while(0)

/* zero an array */
#define array_zero(arr, n, i) \
  for(i = 0; i < n; ++i) { arr[i] = 0; }

#define array_copy(src, dest, n, i) \
  for(i = 0; i < n; ++i) { dest[i] = src[i]; }

#define inner_product(x, y, n, i, result) \
  result = 0; \
  for(i = 0; i < n; ++i) { result += x[i]*y[i]; }

/* find the difference between two struct timevals in seconds */
#define tvdiff(x,y) \
 (((x).tv_sec - (y).tv_sec) + \
  (double)((x).tv_usec - (y).tv_usec)/1000000.0)

/* get a random number in [0,1] */
#define drand() (((double)random())/RAND_MAX)

/* get a random number in [0,1) -- note half-open.  note that we don't
   do (double)(RAND_MAX + 1) because of the risk of overflow */
#define drand2() (((double)random())/((double)(RAND_MAX)+1.0))

/* pick a random integer in [0,max]. */
#define irand(max) (int)(drand2()*(max+1))

#define swap(x, y, tmp) tmp = x; x = y; y = tmp;

/* randomly reorder an array */
#define array_randomize(arr, N, i, r, tmp) \
  for(i = N - 1; i > 0; --i) { \
    r = irand(i); \
    swap(arr[r], arr[i], tmp); \
  }

/* advance len until it equals the length */
#define macro_strlen(str, len) \
  for(len = 0; str[len] != '\0'; ++len) {}

/* first advance len to find the length, then copy the string
   backwards from len */
#define duplicate_string(src, dest, len) \
  macro_strlen(src, len); \
  dest = (char*)malloc((len + 1)*sizeof(char)); \
  for(; len >= 0; --len) { dest[len] = src[len]; }


#endif /* MACROS_H */
